#include "tcpdispatcher"

static map <unsigned long, queue <time_t> > accesslog;
static time_t accesslog_lastclean = 0;

void TcpDispatcher::execute() {
    msg ((Mstr("Dispatch request for client fd ") + clientfd()) + "\n");

    // Check 'softmaxconnrate' and 'hardmaxconnrate' now!
    // Descend into this block if connrate_time() is set, AND
    // either hardmaxconnrate() is set,
    // or both softmaxconnrate() and defertime() are set.
    if (config.connrate_time() &&
	(config.hardmaxconnrate() ||
	 (config.softmaxconnrate() && config.defertime()))) {
     	time_t now, min_ts;
     	now = time(0);
     	min_ts = now - config.connrate_time();
     	unsigned max_conns = max(config.hardmaxconnrate(),
     	  			 config.softmaxconnrate());

	Mutex::lock (&accesslog[client_ip.s_addr]);
	accesslog[client_ip.s_addr].push(now);
	Mutex::unlock (&accesslog[client_ip.s_addr]);
	
	if (accesslog_lastclean < min_ts) {
	    // Clean the entire access log, it's been a while...

	    Mutex::lock(&accesslog_lastclean);
	    accesslog_lastclean = now;
	    Mutex::unlock(&accesslog_lastclean);

	    for (map<unsigned long, queue <time_t> >::iterator i =
		     accesslog.begin();
		 i != accesslog.end();
		 i++ ) {
		if (accesslog[i->first].back() < min_ts) {
		    // This IP hasn't made ANY connections in a while -- erase!
		    accesslog.erase(i);
		} else {
		    // Keep popping off this IP's oldest connection until we
		    // have only "recent" connections left.
		    Mutex::lock(&accesslog[i->first]);
		    while ( accesslog[i->first].front() < min_ts
			    || accesslog[i->first].size() > max_conns ) {
		    	accesslog[i->first].pop();
		    }
		    Mutex::unlock(&accesslog[i->first]);
		}	
	    }
	} else {
	    // The "big log" doesn't need to be fully cleaned,
	    // but this particular IP should be!
	    Mutex::lock(&accesslog[client_ip.s_addr]);
	    while ( accesslog[client_ip.s_addr].front() < min_ts
		    || accesslog[client_ip.s_addr].size() > max_conns ) {
	     	accesslog[client_ip.s_addr].pop();
	    }
	    Mutex::unlock(&accesslog[client_ip.s_addr]);
	}

	if (config.hardmaxconnrate() &&
	    accesslog[client_ip.s_addr].size() >= config.hardmaxconnrate() ) {
	    // This IP has violated the "HARD" limit!  Reject the connection
	    ostringstream o;
	    o << "Client " << inet_ntoa(client_ip)
	      << " has hit the HARD maximum number of connections ("
	      << config.hardmaxconnrate() << " conections in "
	      << config.connrate_time() << " seconds; "
	      << accesslog[client_ip.s_addr].size()
	      << " connections recorded). Client is refused.\n";
	    warnmsg (o.str());
	    socketclose(clientfd());
	    return;
	} else if (config.softmaxconnrate() &&
		   (accesslog[client_ip.s_addr].size() >=
		    config.softmaxconnrate())) {
	    // This IP has violated the "SOFT" Limit. Go to sleep for a while.
	    ostringstream o;
	    o << "Client " << inet_ntoa(client_ip)
	      << " has hit the SOFT maximum number of connections ("
	      << config.softmaxconnrate() << " connections in "
	      << config.connrate_time() << " sedonds; "
	      << accesslog[client_ip.s_addr].size()
	      << " connections recorded). Client is deferred for "
	      << config.defertime() << " microseconds.\n";
	    warnmsg (o.str());
	    usleep(config.defertime());
	}
    }

    try {
	dispatch();
    } catch (Error const &e) {
	cerr << e.what() << "\n";
	socketclose (clientfd());
	return;
    }

    msg ((Mstr("Dispatchign client fd ") + clientfd()) +
	 (Mstr(" to ") + balancer.backend(target_backend).description()) +
	 (Mstr(", fd ") + backendfd()) + "\n");

    balancer.backend(target_backend).startconnection();

    try {
	handle();
    } catch (Error const &e) {
	cerr << e.what() << "\n";
    }

    balancer.backend(target_backend).endconnection();

    socketclose (clientfd());
    socketclose (backendfd());

    msg ((Mstr("Done dispatching to back end fd ") + backendfd()) +
	 (Mstr(" at ") + balancer.backend(target_backend).description()) +
	 "\n");
}
