#include "webinterface"

static unsigned str2uns (string const &s, string const &desc) {
    unsigned ret;

    if (sscanf (s.c_str(), "%u", &ret) < 1)
	throw static_cast<Error>("Bad ") + desc;
    return (ret);
}

static double str2dbl (string const &s, string const &desc) {
    double ret;

    if (sscanf (s.c_str(), "%lf", &ret) < 0)
	throw static_cast<Error>("Bad ") + desc;
    return (ret);
}

static unsigned backendindex (string const &s) {
    unsigned ret;

    ret = str2uns (s, "back end index");
    if (ret >= balancer.nbackends())
	throw static_cast<Error>("Back end index out of range");
    return (ret);
}

static unsigned headerindex (string const &s) {
    unsigned ret;

    ret = str2uns (s, "header index");
    if (ret >= config.nserverheaders())
	throw static_cast<Error>("Server header index out of range");
    return (ret);
}

bool str2bool (string const &s, string const &desc) {
    int i;
    bool ret;

    if (sscanf (s.c_str(), "%d", &i) > 0)
	ret = (i != 0);
    else if (s == "on")
	ret = true;
    else if (s == "off")
	ret = false;
    else
	throw static_cast<Error>("Bad ") + desc + " switch '" + s + "'";

    return (ret);
}

string decode (string const &s) {
    string ret;

    for (char const *cp = s.c_str(); cp && *cp;) {
	if (*cp == '%') {
	    int v;
	    cp++;
	    if (sscanf (cp, "%2x", &v)) {
		ret += static_cast<char>(v);
		cp += 2;
	    }else {
		ret += '%';
	    }
	} else if (*cp == '+') {
	    ret += ' ';
	} else {
	    ret += *cp;
	    cp++;
	}
    }

    // debugmsg ("Decoded: '" + s + "' into '" + ret + "'\n");

    return (ret);
}

void Webinterface::answer(Httpbuffer req) {
    if (req.requestmethod() != Httpbuffer::m_get)
	throw static_cast<Error>("Only request method GET supported");

    string uri = req.requesturi();

    // Status overview
    if (uri == "/") {
	answer_status();
	return;
    }

    // XSLT request
    if (uri == "/xslt") {
	answer_xslt();
	return;
    }

    vector<string> parts = str2parts (uri, '/');
    for (unsigned i = 0; i < parts.size(); i++)
	parts[i] = decode(parts[i]);

    // server/buffersize/VALUE
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "buffersize") {
	unsigned sz = str2uns (parts[2], "buffer size");
	if (sz < 1)
	    throw static_cast<Error>("Buffer size may not be less than 1");
	config.buffersize(sz);
	answer_status();
	return;
    }

    // /server/maxconnections/
    // /server/maxconnections/NUMBER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "maxconnections") ) {
	unsigned num = 0;
	if (parts.size() == 3)
	    num = str2uns (parts[2], "server weight");
	config.maxconn(num);
	answer_status();
	return;
    }

    // /server/addxrversion/BOOLEAN
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "addxrversion") {
	config.addxrversion (str2bool (parts[2], "addxrversion"));
	answer_status();
	return;
    }

    // /server/addxforwardedfor/BOOLEAN
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "addxforwardedfor") {
	config.addxforwardedfor (str2bool (parts[2], "addxforwardedfor"));
	answer_status();
	return;
    }

    // /server/stickyhttp/BOOLEAN
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "stickyhttp") {
	config.stickyhttp (str2bool(parts[2], "stickyhttp"));
	answer_status();
	return;
    }

    // /server/newheader/NEWHEADER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "newheader") ) {
	if (parts.size() == 3)
	    config.addserverheader(parts[2]);
	answer_status();
	return;
    }

    // /server/changeheader/NR
    // /server/changeheader/NR/VALUE
    if ( (parts.size() == 3 || parts.size() == 4) &&
	 (parts[0] == "server" && parts[1] == "changeheader") ) {
	unsigned ind = headerindex(parts[2]);
	if (parts.size() == 3)
	    config.removeserverheader(ind);
	else
	    config.changeserverheader(ind, parts[3]);
	answer_status();
	return;
    }

    // /server/verbose/BOOLEAN
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "verbose") {
	config.verbose(str2bool(parts[2], "verbose"));
	answer_status();
	return;
    }

    // /server/debug/VERBOSE
    if (parts.size() == 3 &&
	parts[0] == "server" && parts[1] == "debug") {
	config.debug(str2bool(parts[2], "debug"));
	answer_status();
	return;
    }

    // /server/logtrafficdir
    // /server/logtrafficdir/VALUE
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "logtrafficdir") ) {
	if (parts.size() == 2)
	    config.dumpdir("");
	else
	    config.dumpdir(parts[2]);
	answer_status();
	return;
    }

    // /server/clienttimeout
    // /server/clienttimeout/NUMBER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "clienttimeout") ) {
	unsigned num = 0;
	if (parts.size() == 3)
	    num = str2uns (parts[2], "client timeout");
	config.client_timeout(num);
	answer_status();
	return;
    }

    // /server/backendtimeout
    // /server/backendtimeout/NUMBER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "backendtimeout") ) {
	unsigned num = 0;
	if (parts.size() == 3)
	    num = str2uns (parts[2], "client timeout");
	config.backend_timeout(num);
	answer_status();
	return;
    }

    // /server/wakeupinterval
    // /server/wakeupinterval/NUMBER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "wakeupinterval") ) {
	unsigned num = 0;
	if (parts.size() == 3)
	    num = str2uns (parts[2], "wakeup interval");
	if (num)
	    config.checkupsec(0);
	config.wakeupsec(num);
	answer_status();
	return;
    }

    // /server/checkupinterval
    // /server/checkupinterval/NUMBER
    if ( (parts.size() == 2 || parts.size() == 3) &&
	 (parts[0] == "server" && parts[1] == "checkupinterval") ) {
	unsigned num = 0;
	if (parts.size() == 3)
	    num = str2uns (parts[2], "checkup interval");
	if (num)
	    config.wakeupsec(0);
	config.checkupsec(num);
	answer_status();
	return;
    }

    // /backend/NR/weight/NUMBER
    if (parts.size() == 4 &&
	parts[0] == "backend" && parts[2] == "weight") {
	unsigned ind = backendindex(parts[1]);
	unsigned num = str2uns (parts[3], "back end weight");
	if (num < 1)
	    throw static_cast<Error>("Weight may not be less than 1");
	balancer.backend(ind).weight(num);
	answer_status();
	return;
    }

    // /backend/NR/maxconnections/NUMBER
    if (parts.size() == 4 &&
	parts[0] == "backend" && parts[2] == "maxconnections") {
	unsigned ind = backendindex(parts[1]);
	unsigned num = str2uns (parts[3], "back end maxconnections");
	balancer.backend(ind).maxconn(num);
	answer_status();
	return;
    }

    // /backend/NR/loadavg/FLOAT
    if (parts.size() == 4 &&
	parts[0] == "backend" && parts[2] == "loadavg") {
	unsigned ind = backendindex(parts[1]);
	double fnum = str2dbl (parts[3], "back end loadavg");
	balancer.backend(ind).loadavg(fnum);
	answer_status();
	return;
    }

    // /backend/NR/hostmatch/EXPRESSION
    // /backend/NR/hostmatch
    if ( (parts.size() == 3 || parts.size() == 4) &&
	 (parts[0] == "backend" && parts[2] == "hostmatch") ) {
	unsigned ind = backendindex(parts[1]);
	balancer.backend(ind).hostmatch(parts.size() == 3 ? "" : parts[3]);
	answer_status();
	return;
    }

    // /backend/NR/up/BOOL
    if (parts.size() == 4 && parts[0] == "backend" && parts[2] == "up") {
	unsigned ind = backendindex(parts[1]);
	balancer.backend(ind).up(str2bool(parts[3], "up"));
	answer_status();
	return;
    }

    throw static_cast<Error>("No action for URI '") + uri + "'";
}
